@@ -71,6 +71,7 @@ import com.openexchange.sessiond.cache.SessionCache;
 import com.openexchange.sessiond.exception.SessiondException;
 import com.openexchange.sessiond.exception.SessiondException.Code;
 import com.openexchange.sessiond.services.SessiondServiceRegistry;
+import com.openexchange.threadpool.ThreadPoolService;
 import com.openexchange.timer.ScheduledTimerTask;
 import com.openexchange.timer.TimerService;
 
@@ -187,14 +188,14 @@ public final class SessionHandler {
      * @return The session ID associated with newly created session
      * @throws SessiondException If creating a new session fails
      */
-    protected static String addSession(final int userId, final String loginName, final String password, final Context context, final String clientHost, final String login, String authId) throws SessiondException {
+    protected static String addSession(final int userId, final String loginName, final String password, final Context context, final String clientHost, final String login, String authId, String hash) throws SessiondException {
         checkMaxSessPerUser(userId, context);
         checkAuthId(login, authId);
         final String sessionId = sessionIdGenerator.createSessionId(loginName, clientHost);
         final Session session =
             new SessionImpl(userId, loginName, password, context.getContextId(), sessionId, sessionIdGenerator.createSecretId(
                 loginName,
-                Long.toString(System.currentTimeMillis())), sessionIdGenerator.createRandomId(), clientHost, login, authId);
+                Long.toString(System.currentTimeMillis())), sessionIdGenerator.createRandomId(), clientHost, login, authId, hash);
         /*
          * Add session
          */
@@ -282,24 +283,22 @@ public final class SessionHandler {
     /**
      * Gets the session associated with given session ID
      * 
-     * @param sessionid The session ID
+     * @param sessionId The session ID
      * @return The session associated with given session ID; otherwise <code>null</code> if expired or none found
      */
-    protected static SessionControl getSession(final String sessionid) {
+    protected static SessionControl getSession(String sessionId) {
         if (DEBUG) {
-            LOG.debug(new StringBuilder("getSession <").append(sessionid).append('>').toString());
+            LOG.debug(new StringBuilder("getSession <").append(sessionId).append('>').toString());
         }
-        final SessionControl sessionControl = sessionData.getSession(sessionid);
+        final SessionControl sessionControl = sessionData.getSession(sessionId);
         if (null == sessionControl) {
             return null;
         }
-        /*
-         * Look-up cache if current session wrapped by session-control is marked for removal
-         */
+        // Look-up cache if current session wrapped by session-control is marked for removal
         try {
             final SessionCache cache = SessionCache.getInstance();
-            final Session s = sessionControl.getSession();
-            final CachedSession cachedSession = cache.getCachedSessionByUser(s.getUserId(), s.getContextId());
+            final Session session = sessionControl.getSession();
+            final CachedSession cachedSession = cache.getCachedSessionByUser(session.getUserId(), session.getContextId());
             if (null != cachedSession) {
                 if (cachedSession.isMarkedAsRemoved()) {
                     cache.removeCachedSession(cachedSession.getSecret());
@@ -316,15 +315,14 @@ public final class SessionHandler {
     }
 
     /**
-     * Gets (and removes) the session bound to given secret cookie identifier in cache.
+     * Gets (and removes) the session bound to given session identifier in cache.
      * <p>
      * Session is going to be added to local session containers on a cache hit.
      * 
      * @param sessionId The session identifier
-     * @param localIP The host's local IP
      * @return A wrapping instance of {@link SessionControl} or <code>null</code>
      */
-    public static SessionControl getCachedSession(final String sessionId, final String localIP) {
+    public static SessionControl getCachedSession(final String sessionId) {
         if (DEBUG) {
             LOG.debug(new StringBuilder("getCachedSession <").append(sessionId).append('>').toString());
         }
@@ -335,8 +333,8 @@ public final class SessionHandler {
                     removeUserSessions(cachedSession.getUserId(), cachedSession.getContextId(), false);
                 } else {
                     // A cache hit! Add to local session containers
-                    LOG.info("Cached session found. ID: " + cachedSession.getSessionId());
-                    return sessionData.addSession(new SessionImpl(cachedSession, localIP), config.getLifeTime(), noLimit);
+                    LOG.info("Migrate session: " + cachedSession.getSessionId());
+                    return sessionData.addSession(new SessionImpl(cachedSession), config.getLifeTime(), noLimit);
                 }
             }
         } catch (final CacheException e) {
@@ -435,4 +433,12 @@ public final class SessionHandler {
             }
         }
     }
+
+    public static void addThreadPoolService(ThreadPoolService service) {
+        sessionData.addThreadPoolService(service);
+    }
+
+    public static void removeThreadPoolService() {
+        sessionData.removeThreadPoolService();
+    }
 }
