@@ -51,7 +51,6 @@ package com.openexchange.sessiond.impl;
 
 import java.util.UUID;
 import com.openexchange.java.util.UUIDs;
-import com.openexchange.sessiond.exception.SessiondException;
 
 /**
  * {@link UUIDSessionIdGenerator} - The session ID generator based on {@link UUID#randomUUID()}.
@@ -68,17 +67,17 @@ public final class UUIDSessionIdGenerator extends SessionIdGenerator {
     }
 
     @Override
-    public String createSessionId(final String userId, final String data) throws SessiondException {
+    public String createSessionId(final String userId, final String data) {
         return randomUUID();
     }
 
     @Override
-    public String createSecretId(final String userId, final String data) throws SessiondException {
+    public String createSecretId(final String userId, final String data) {
         return randomUUID();
     }
 
     @Override
-    public String createRandomId() throws SessiondException {
+    public String createRandomId() {
         return randomUUID();
     }
 
